@@ -124,6 +124,7 @@ qw(
 	package_multiarch
 	package_section
 	package_arch
+	package_type
 	process_pkg
 	compute_doc_main_package
 	isnative
@@ -133,6 +134,7 @@ qw(
 qw(
 	basename
 	dirname
+	mkdirs
 	install_file
 	install_prog
 	install_lib
@@ -227,6 +229,7 @@ our $PKGVERSION_REGEX = qr/
                  (?: - [0-9A-Za-z.+:~]+ )*   # Optional debian revision (+ upstreams versions with hyphens)
                           /xoa;
 our $MAINTSCRIPT_TOKEN_REGEX = qr/[A-Za-z0-9_.+]+/o;
+our $TOOL_NAME = basename($0);
 
 # From Policy 5.1:
 #
@@ -374,7 +377,7 @@ sub END {
 	# expect.
 	return if not -f 'debian/control';
 	if (compat(9, 1) || $ENV{DH_INTERNAL_OVERRIDE}) {
-		write_log(basename($0), @{$dh{DOPACKAGES}});
+		write_log($TOOL_NAME, @{$dh{DOPACKAGES}});
 	}
 }
 
@@ -486,6 +489,35 @@ sub print_and_doit_noerror {
 	goto \&_doit;
 }
 
+sub _post_fork_setup_and_exec {
+	my ($close_stdin, $options, @cmd) = @_;
+	if (defined($options)) {
+		if (defined(my $dir = $options->{chdir})) {
+			if ($dir ne '.') {
+				chdir($dir) or error("chdir(\"${dir}\") failed: $!");
+			}
+		}
+		if ($close_stdin) {
+			open(STDIN, '<', '/dev/null') or error("redirect STDIN failed: $!");
+		}
+		if (defined(my $output = $options->{stdout})) {
+			open(STDOUT, '>', $output) or error("redirect STDOUT failed: $!");
+		}
+		if (defined(my $update_env = $options->{update_env})) {
+			while (my ($k, $v) = each(%{$update_env})) {
+				if (defined($v)) {
+					$ENV{$k} = $v;
+				} else {
+					delete($ENV{$k});
+				}
+			}
+		}
+	}
+	# Force execvp call to avoid shell.  Apparently, even exec can
+	# involve a shell if you don't do this.
+	exec { $cmd[0] } @cmd or error('exec (for cmd: ' . escape_shell(@cmd) . ") failed: $!");
+}
+
 sub _doit {
 	my (@cmd) = @_;
 	my $options = ref($cmd[0]) ? shift(@cmd) : undef;
@@ -499,29 +531,7 @@ sub _doit {
 	return 1 if $dh{NO_ACT};
 	my $pid = fork() // error("fork(): $!");
 	if (not $pid) {
-		if (defined($options)) {
-			if (defined(my $dir = $options->{chdir})) {
-				if ($dir ne '.') {
-					chdir($dir) or error("chdir(\"${dir}\") failed: $!");
-				}
-			}
-			open(STDIN, '<', '/dev/null') or error("redirect STDIN failed: $!");
-			if (defined(my $output = $options->{stdout})) {
-				open(STDOUT, '>', $output) or error("redirect STDOUT failed: $!");
-			}
-			if (defined(my $update_env = $options->{update_env})) {
-				while (my ($k, $v) = each(%{$update_env})) {
-					if (defined($v)) {
-						$ENV{$k} = $v;
-					} else {
-						delete($ENV{$k});
-					}
-				}
-			}
-		}
-		# Force execvp call to avoid shell.  Apparently, even exec can
-		# involve a shell if you don't do this.
-		exec { $cmd[0] } @cmd;
+		_post_fork_setup_and_exec(1, $options, @cmd) // error("Assertion error: sub should not return!");
 	}
 	return waitpid($pid, 0) == $pid && $? == 0;
 }
@@ -560,8 +570,12 @@ sub _format_cmdline {
 
 sub qx_cmd {
 	my (@cmd) = @_;
+	my $options = ref($cmd[0]) ? shift(@cmd) : undef;
 	my ($output, @output);
-	open(my $fd, '-|', @cmd) or error('fork+exec (' . escape_shell(@cmd) . "): $!");
+	my $pid = open(my $fd, '-|') // error('fork (for cmd: ' . escape_shell(@cmd) . ") failed: $!");
+	if ($pid == 0) {
+		_post_fork_setup_and_exec(0, $options, @cmd) // error("Assertion error: sub should not return!");
+	}
 	if (wantarray) {
 		@output = <$fd>;
 	} else {
@@ -655,28 +669,42 @@ sub error_exitcode {
 	}
 }
 
-sub install_dir {
-	my @to_create = grep { not -d $_ } @_;
-	return if not @to_create;
+
+sub _mkdirs {
+	my ($log, @dirs) = @_;
+	return if not @dirs;
+	if ($log && $dh{VERBOSE}) {
+		verbose_print(sprintf('install -m0755 -d %s', escape_shell(@dirs)));
+	}
+	return 1 if $dh{NO_ACT};
 	state $_loaded;
 	if (not $_loaded) {
 		$_loaded++;
 		require File::Path;
 	}
-	verbose_print(sprintf('install -d %s', escape_shell(@to_create)))
-		if $dh{VERBOSE};
-	return 1 if $dh{NO_ACT};
+	my %opts = (
+		# install -d uses 0755 (no umask), make_path uses 0777 (& umask) by default.
+		# Since we claim to run install -d, then ensure the mode is correct.
+		'chmod' => 0755,
+	);
 	eval {
-		File::Path::make_path(@to_create, {
-			# install -d uses 0755 (no umask), make_path uses 0777 (& umask) by default.
-			# Since we claim to run install -d, then ensure the mode is correct.
-			'chmod' => 0755,
-		});
+		File::Path::make_path(@dirs, \%opts);
 	};
 	if (my $err = "$@") {
 		$err =~ s/\s+at\s+\S+\s+line\s+\d+\.?\n//;
 		error($err);
 	}
+	return;
+}
+
+sub mkdirs {
+	my @to_create = grep { not -d $_ } @_;
+	return _mkdirs(0, @to_create);
+}
+
+sub install_dir {
+	my @dirs = @_;
+	return _mkdirs(1, @dirs);
 }
 
 sub rename_path {
@@ -849,7 +877,7 @@ sub error {
 	my ($message) = @_;
 	# ensure the error code is well defined.
 	$! = 255;
-	die(_color(basename($0), 'bold') . ': ' . _color('error', 'bold red') . ": $message\n");
+	die(_color($TOOL_NAME, 'bold') . ': ' . _color('error', 'bold red') . ": $message\n");
 }
 
 # Output a warning.
@@ -857,7 +885,7 @@ sub warning {
 	my ($message) = @_;
 	$message //= '';
 
-	print STDERR _color(basename($0), 'bold') . ': ' . _color('warning', 'bold yellow') . ": $message\n";
+	print STDERR _color($TOOL_NAME, 'bold') . ': ' . _color('warning', 'bold yellow') . ": $message\n";
 }
 
 # Returns the basename of the argument passed to it.
@@ -1230,14 +1258,13 @@ sub autoscript {
 		}
 	}
 
-	if (-e $outfile && ($script eq 'postrm' || $script eq 'prerm')
-	   && !compat(5)) {
+	if (-e $outfile && ($script eq 'postrm' || $script eq 'prerm')) {
 		# Add fragments to top so they run in reverse order when removing.
 		if (not defined($sed) or ref($sed)) {
 			verbose_print("[META] Prepend autosnippet \"$filename\" to $script [${outfile}.new]");
 			if (not $dh{NO_ACT}) {
 				open(my $out_fd, '>', "${outfile}.new") or error("open(${outfile}.new): $!");
-				print {$out_fd} '# Automatically added by ' . basename($0) . "/${tool_version}\n";
+				print {$out_fd} '# Automatically added by ' . $TOOL_NAME . "/${tool_version}\n";
 				autoscript_sed($sed, $infile, undef, $out_fd);
 				print {$out_fd} "# End automatically added section\n";
 				open(my $in_fd, '<', $outfile) or error("open($outfile): $!");
@@ -1248,7 +1275,7 @@ sub autoscript {
 				close($out_fd) or error("close(${outfile}.new): $!");
 			}
 		} else {
-			complex_doit("echo \"# Automatically added by ".basename($0)."/${tool_version}\"> $outfile.new");
+			complex_doit("echo \"# Automatically added by ".$TOOL_NAME."/${tool_version}\"> $outfile.new");
 			autoscript_sed($sed, $infile, "$outfile.new");
 			complex_doit("echo '# End automatically added section' >> $outfile.new");
 			complex_doit("cat $outfile >> $outfile.new");
@@ -1258,13 +1285,13 @@ sub autoscript {
 		verbose_print("[META] Append autosnippet \"$filename\" to $script [${outfile}]");
 		if (not $dh{NO_ACT}) {
 			open(my $out_fd, '>>', $outfile) or error("open(${outfile}): $!");
-			print {$out_fd} '# Automatically added by ' . basename($0) . "/${tool_version}\n";
+			print {$out_fd} '# Automatically added by ' . $TOOL_NAME . "/${tool_version}\n";
 			autoscript_sed($sed, $infile, undef, $out_fd);
 			print {$out_fd} "# End automatically added section\n";
 			close($out_fd) or error("close(${outfile}): $!");
 		}
 	} else {
-		complex_doit("echo \"# Automatically added by ".basename($0)."/${tool_version}\">> $outfile");
+		complex_doit("echo \"# Automatically added by ".$TOOL_NAME."/${tool_version}\">> $outfile");
 		autoscript_sed($sed, $infile, $outfile);
 		complex_doit("echo '# End automatically added section' >> $outfile");
 	}
@@ -1333,7 +1360,7 @@ sub autoscript_sed {
                               }x;
 			print {$ofd} $line;
 		}
-		print {$ofd} '# Triggers added by ' . basename($0) . "/${tool_version}\n";
+		print {$ofd} '# Triggers added by ' . $TOOL_NAME . "/${tool_version}\n";
 		print {$ofd} "${trigger_type} ${trigger_target}\n";
 		close($ofd) or error("closing ${triggers_file}.new failed: $!");
 		close($ifd);
@@ -1350,7 +1377,7 @@ sub generated_file {
 	my $dir = "debian/.debhelper/generated/${package}";
 	my $path = "${dir}/${filename}";
 	$mkdirs //= 1;
-	install_dir($dir) if $mkdirs;
+	mkdirs($dir) if $mkdirs;
 	return $path;
 }
 
@@ -1372,7 +1399,7 @@ sub _update_substvar {
 	my $len = scalar(@lines);
 	push(@lines, $insert_logic->()) if $insert_logic;
 	$changed ||= $len != scalar(@lines);
-	if ($changed) {
+	if ($changed && !$dh{NO_ACT}) {
 		open(my $out, '>', "${substvar_file}.new") // error("open(${substvar_file}.new, \"w\"): $!");
 		for my $line (@lines) {
 			print {$out} "$line\n";
@@ -1405,6 +1432,19 @@ sub addsubstvar {
 	my $substvarfile = "debian/${ext}substvars";
 	my $str = $deppackage;
 	$str .= " ($verinfo)" if defined $verinfo && length $verinfo;
+
+	if (not defined($deppackage) and not $remove) {
+		error("Bug in helper: Must provide a value for addsubstvar (or set the remove flag, but then use delsubstvar instead)")
+	}
+
+	if (defined($str) and $str =~ m/[\n]/) {
+		$str =~ s/\n/\\n/g;
+		# Per #1026014
+		warning('Unescaped newlines in the value of a substvars can cause broken substvars files (see #1025714).');
+		warning("Hint: If you really need a newline character, provide it as \"\${Newline}\".");
+		error("Bug in helper: The substvar must not contain a raw newline character (${substvar}=${str})");
+	}
+
 	my $update_logic = sub {
 		my ($line) = @_;
 		return $line if $line !~ m/^\Q${substvar}\E([?]?=)(.*)/;
@@ -2165,17 +2205,24 @@ sub package_cross_type {
 	return $package_cross_type{$package} // 'host';
 }
 
+sub package_type {
+	my ($package) = @_;
+
+	if (! exists $package_types{$package}) {
+		warning "package $package is not in control info";
+		return DEFAULT_PACKAGE_TYPE;
+	}
+	return $package_types{$package};
+}
+
 # Return true if a given package is really a udeb.
 sub is_udeb {
 	my $package=shift;
 	
-	if (! exists $package_types{$package}) {
-		warning "package $package is not in control info";
-		return 0;
-	}
-	return $package_types{$package} eq 'udeb';
+	return package_type($package) eq 'udeb';
 }
 
+
 sub process_pkg {
 	my ($package) = @_;
 	state %packages_to_process = map { $_ => 1 } @{$dh{DOPACKAGES}};
@@ -2560,7 +2607,7 @@ sub setup_home_and_xdg_dirs {
 		XDG_DATA_DIRS
 		XDG_RUNTIME_DIR
 	);
-	install_dir(@paths);
+	mkdirs(@paths);
 	for my $envname (@clear_env) {
 		delete($ENV{$envname});
 	}
@@ -2627,7 +2674,7 @@ sub get_buildoption {
 		} elsif ($opt =~ m/^dherroron=(\S*)$/ && $wanted eq 'dherroron') {
 			my $value = $1;
 			if ($value ne 'obsolete-compat-levels') {
-				warning("Unknown value \"${value}\" as parameter for \"dherrron\" seen in DEB_BUILD_OPTIONS");
+				warning("Unknown value \"${value}\" as parameter for \"dherroron\" seen in DEB_BUILD_OPTIONS");
 			}
 			return $value;
 		} elsif ($opt eq $wanted) {
@@ -2656,7 +2703,7 @@ sub _executable_dh_config_file_failed {
 	# The interpreter did not like the file for some reason.
 	# Lets check if the maintainer intended it to be
 	# executable.
-	if (not is_so_or_exec_elf_file($source) and not _has_shbang_line($source)) {
+	if (not is_so_or_exec_elf_file($source) and not _has_shebang_line($source)) {
 		warning("${source} is marked executable but does not appear to an executable config.");
 		warning();
 		warning("If ${source} is intended to be an executable config file, please ensure it can");
@@ -2676,15 +2723,14 @@ sub _executable_dh_config_file_failed {
 # the package.  Under compat 9+ it may execute the file and use its
 # output instead.
 #
-# install_dh_config_file(SOURCE, TARGET[, MODE])
+# install_dh_config_file(SOURCE, TARGET)
 sub install_dh_config_file {
-	my ($source, $target, $mode) = @_;
-	$mode = 0644 if not defined($mode);
+	my ($source, $target) = @_;
 
 	if (!compat(8) and -x $source) {
 		my @sstat = stat(_) || error("cannot stat $source: $!");
 		open(my $tfd, '>', $target) || error("cannot open $target: $!");
-		chmod($mode, $tfd) || error("cannot chmod $target: $!");
+		chmod(0644, $tfd) || error("cannot chmod $target: $!");
 		open(my $sfd, '-|', $source) || error("cannot run $source: $!");
 		while (my $line = <$sfd>) {
 			print ${tfd} $line;
@@ -2696,7 +2742,7 @@ sub install_dh_config_file {
 		# Set the mtime (and atime) to ensure reproducibility.
 		utime($sstat[9], $sstat[9], $target);
 	} else {
-		_install_file_to_path($mode, $source, $target);
+		install_file($source, $target);
 	}
 	return 1;
 }
@@ -2706,7 +2752,7 @@ sub restore_file_on_clean {
 	my $bucket_index = 'debian/.debhelper/bucket/index';
 	my $bucket_dir = 'debian/.debhelper/bucket/files';
 	my $checksum;
-	install_dir($bucket_dir);
+	mkdirs($bucket_dir);
 	if ($file =~ m{^/}) {
 		error("restore_file_on_clean requires a path relative to the package dir");
 	}
@@ -2814,8 +2860,14 @@ sub log_installed_files {
 	my ($package, @patterns) = @_;
 
 	return if $dh{NO_ACT};
+	my $tool = $TOOL_NAME;
+	if (ref($package) eq 'HASH') {
+		my $options = $package;
+		$tool = $options->{'tool_name'} // error('Missing mandatory "tool_name" option for log_installed_files');
+		$package = $options->{'package'} // error('Missing mandatory "package" option for log_installed_files');
+	}
 
-	my $log = generated_file($package, 'installed-by-' . basename($0));
+	my $log = generated_file($package, 'installed-by-' . $tool);
 	open(my $fh, '>>', $log) or error("open $log: $!");
 	for my $src (@patterns) {
 		print $fh "$src\n";
@@ -2871,7 +2923,7 @@ sub is_so_or_exec_elf_file {
 	return 1;
 }
 
-sub _has_shbang_line {
+sub _has_shebang_line {
 	my ($file) = @_;
 	open(my $fd, '<', $file) or error("open $file: $!");
 	my $line = <$fd>;
